from dataclasses import dataclass
from typing import Union

@dataclass
class Entry:
    categories: list[str]
    is_income: bool
    is_calculated: bool
    amount: int
    children: list[str]
 
@dataclass
class ReportData:
    entries: list[Entry]
    total_income: int
    income_dict: dict[str, Entry]
    total_expenses: int
    expense_dict: dict[str, Entry]
    total_unassigned: int

@dataclass 
class DiffTreeNode:
    category_name: str
    category_totals: list[int] 
    children: dict[str, "DiffTreeNode"]



from minibudget import parse
from minibudget import render
from minibudget import transform
from minibudget import convert
from minibudget.render import RenderOptions
from rich.console import Console
from pathlib import Path

class CommonParser:
    @staticmethod
    def setup_render_options(parser):
        parser.add_argument("--width", help="Width of the rendered report in characters. Defaults to the space available in the terminal.", type=int)
        parser.add_argument("--currency", 
                            type=str, 
                            help="The currency to render this budget with. A shortcut for --currency-format and --currency-decimals")
        parser.add_argument("--currency-format", 
                            default="{neg}${amount}", 
                            help="Currency format, using Python format string syntax. E.g. {neg}${amount}")
        parser.add_argument("--currency-decimals", 
                            type=int, 
                            default=2, 
                            help="Number of decimal places to display when rendering currency. E.g. 2 will render as $0.00, while 0 will render as $0.")
    
    @staticmethod
    def get_render_options(args) -> RenderOptions: 
        if args.currency_decimals < 0:
            raise ValueError("Currency decimals must be 0 or more.")
        
        render_data = RenderOptions(
                    args.width,
                    args.currency_format,
                    args.currency_decimals
                )

        if args.currency in render.PREDEFINED_CURRENCIES:
            currency_data = render.PREDEFINED_CURRENCIES[args.currency]
            render_data.currency_format = currency_data.currency_format
            render_data.currency_decimals = currency_data.currency_decimals
        
        return render_data

class ReportParser:
    @staticmethod
    def setup(parent_subparser):
        report_parser = parent_subparser.add_parser("report", help="Report on a single .budget file.")
        CommonParser.setup_render_options(report_parser)
        report_parser.add_argument("file")
        report_parser.set_defaults(func=ReportParser.report)

    @staticmethod
    def report(args): 
        entries = parse.budget(args.file)
        
        report_data = transform.entries_to_report_data(entries)
        render_data = CommonParser.get_render_options(args)

        render.report(report_data, render_data)

class DiffParser:
    @staticmethod
    def setup(parent_subparser):
        diff_parser = parent_subparser.add_parser("diff", help="See the difference between each category in several .budget files. Each file is considered one time period and differences are rolling between periods.")
        CommonParser.setup_render_options(diff_parser)
        diff_parser.add_argument("files", nargs="+")
        diff_parser.set_defaults(func=DiffParser.diff)

    @staticmethod
    def diff(args):
        render_data = CommonParser.get_render_options(args)
        if len(args.files) < 2:
            raise ValueError("Must have at least 2 files to produce a diff.")

        file_entries = [ parse.budget(filename) for filename in args.files ]

        category_trees = [ transform.generate_category_dict(f) for f in file_entries]
        diff_tree = transform.generate_diff_dict(category_trees)
        names = [ Path(f).stem for f in args.files ]
        table = render.diff_tree(diff_tree, names, render_data)
        console = Console()
        console.print(table)

class ConvertParser:
    @staticmethod
    def setup(parent_subparser):
        convert_parser = parent_subparser.add_parser("convert", help="Convert to minibudget format from other financial formats.")
        convert_parser.add_argument("file")
        convert_parser.add_argument("--width", help="Width of the output minibudget in characters. Default is 80.", default=80)
        convert_parser.add_argument("--start", help="Start date to query from, inclusive.")
        convert_parser.add_argument("--end", help="End date to query until, inclusive.")
        convert_parser.add_argument("--currency", help="The currency to convert into minibudget format, where multiple are available. Default is USD.", default="USD")
        convert_parser.add_argument("--format", help="Format of the input file to output as minibudget entries.", choices=["beancount"])
        convert_parser.set_defaults(func=ConvertParser.convert)

    @staticmethod
    def convert(args):
        format = ConvertParser.infer_format(args)
        if format == "beancount":
            entries = convert.beancount(args.file, args.currency, args.start, args.end)
        else:
            raise ValueError(f"{args.file} is not a parseable type.")
        print(convert.entry_list_to_string(entries, int(args.width)))
        
    @staticmethod
    def infer_format(args):
        if args.format is not None:
            return args.format
        file_path = Path(args.file)
        if file_path.suffix == ".beancount":
            return "beancount"
        return None
            



import rich
from rich.markup import render
from minibudget.model import ReportData, Entry
from dataclasses import dataclass
from minibudget.helpers import dft_diff_dict, dft_entry_dict
from rich.table import Table
from rich.text import Text
from typing import Union

@dataclass
class RenderOptions:
    width: int
    currency_format: str
    currency_decimals: int

PREDEFINED_CURRENCIES = {
    "NTD": RenderOptions(width=0, currency_format="{neg}{amount} NTD", currency_decimals=0),
    "USD": RenderOptions(width=0, currency_format="{neg}${amount}", currency_decimals=2)
}

def report_table(title: str,categories: dict[str, Entry], total: int, render_data: RenderOptions) -> Table: 
    table = Table(title=title, expand=True)
    if render_data.width != None:
        table.width = render_data.width
        table.expand = False
    table.add_column("Category", ratio=5)
    table.add_column("Amount", justify="right", ratio=2)

    def render_category(entry: Entry):
        depth = len(entry.categories) - 1
        tag = entry.categories[-1]
        left = f"{" " * 4 * depth}{tag}"
        right = currency(entry.amount, render_data)
        table.add_row(left, right)
    
    dft_entry_dict(categories, fn=render_category)
    table.add_section()
    table.add_row("Total", currency(total, render_data))

    return table

def diff_tree(tree: dict[str, list[Union[Entry, None]]], names: list[str], render_data: RenderOptions) -> Table:
    table = Table(expand=True)
    if render_data.width != None:
        table.width = render_data.width
        table.expand = False
    table.add_column("Category")
    for name in names:
        table.add_column(name, justify="right")
 
    def render_category(entries: list[Union[Entry, None]]):
        amounts = []
        for entry in entries:
            if entry == None:
                amounts.append(0)
            else:
                amounts.append(entry.amount)
        tag = ""
        depth = 0 

        for entry in entries:
            if entry != None:
                tag = entry.categories[-1]
                depth = len(entry.categories) - 1

        category = f"{"    "*depth}{tag}"
        cells = [Text( currency(amounts[0], render_data) )]

        for i, amount in enumerate(amounts[1:]):
            diff = amount - amounts[i]
            amount_rendered = f"{currency(amount, render_data)}\n"
            diff_rendered = Text(currency(diff, render_data))
            if diff > 0:
                diff_rendered.stylize("green")
            elif diff == 0:
                diff_rendered.stylize("dim")
            elif diff < 0:
                diff_rendered.stylize("red")
            cells.append( Text.assemble(amount_rendered,diff_rendered) )
        
        if depth == 0:
            table.add_section()
        table.add_row( category, *cells )
        if depth == 0:
            table.add_section()
        
    dft_diff_dict(tree, fn=render_category)
    return table

def currency(units: int, render_data: RenderOptions) -> str:
    # so we can do e.g. -$100 instead of $-100
    amount = str(abs(units))
    decimal = render_data.currency_decimals
    if decimal > 0:
        left = amount[:-decimal]
        if len(left) == 0:
            left = "0"
        right = amount[-decimal:]
        while len(right) < decimal:
            right += "0"
        amount = left + "." + right
    output = render_data.currency_format.format(amount=amount, neg="-" if units < 0 else "") 
    return output

def report(data: ReportData, render_data: RenderOptions):
    console = rich.console.Console()
    
    income_table = report_table("Income",data.income_dict, data.total_income, render_data)
    expense_table = report_table("Expenses",data.expense_dict, data.total_expenses, render_data)

    unassigned_style = "default"
    unassigned_string = "All funds have been assigned. =)"

    if data.total_unassigned < 0:
        unassigned_style = "red"
        unassigned_string = currency(data.total_unassigned, render_data)
    elif data.total_unassigned > 0:
        unassigned_style = "green"
        unassigned_string = currency(data.total_unassigned, render_data)

    unassigned_table = Table(expand=True, show_header=False, border_style=unassigned_style, row_styles=[unassigned_style])
    
    if render_data.width != None:
        unassigned_table.width = render_data.width
        unassigned_table.expand = False

    unassigned_table.add_column(ratio=5)
    unassigned_table.add_column(justify="right", ratio=2)

    unassigned_table.add_row("Unassigned funds", unassigned_string)
 
    console.print(income_table, expense_table, unassigned_table)
 
def diff(reports: list[ReportData], render_data: RenderOptions):
    pass



#!python
from argparse import ArgumentParser
from minibudget.parsers import ReportParser, DiffParser, ConvertParser

def main():
    parser = ArgumentParser()
    subparsers = parser.add_subparsers(required=True)

    ReportParser.setup(subparsers)
    DiffParser.setup(subparsers)
    ConvertParser.setup(subparsers)

    args = parser.parse_args()
    args.func(args) 

if __name__ == "__main__":
    main()



from collections.abc import Callable
from typing import Union
from minibudget.model import Entry

def _dft_entry_dict(entry_dict: dict[str, Entry], root: str, fn: Union[None, Callable]):
    root_entry = entry_dict[root]
    if fn != None:
        fn(root_entry)
    for child in root_entry.children:
        _dft_entry_dict(entry_dict, child, fn)
        
def dft_entry_dict(entry_dict: dict[str, Entry], fn: Union[None, Callable] = None):
    roots = [ key for key in entry_dict.keys() if len(key.split(":")) == 1 ]
    for root in roots:
        _dft_entry_dict( entry_dict, root, fn )

def _dft_diff_dict(entry_dict: dict[str, list[Union[Entry, None]]], root: str, fn: Union[None, Callable]):
    root_list = entry_dict[root]
    if fn != None:
        fn(root_list)
    children = set()
    for entry in root_list:
        if entry == None:
            continue
        children.update(entry.children)
    for child in children:
        _dft_diff_dict(entry_dict, child, fn)

def dft_diff_dict(entry_dict: dict[str, list[Union[Entry, None]]], fn: Union[None, Callable]):
    roots = [ key for key in entry_dict.keys() if len(key.split(":")) == 1 ]
    for root in roots:
        _dft_diff_dict( entry_dict, root, fn )



from typing import Union
from minibudget.model import Entry, ReportData, DiffTreeNode
from copy import deepcopy

def calculate_total(entries: list[Entry]) -> int:
    total = 0

    for entry in entries:        
        to_add = entry.amount
        
        total += to_add

    return total

# By generating categories this way we can render as a tree
# without the algorithmic complexity of storing the information
# as a tree; instead we're using a method similar to a DAG
# representation in the dictionary.
def generate_category_dict(entries: list[Entry]) -> dict[str, Entry]:
    base_output: dict[str, Entry] = {}
    for entry in entries:
        key = ":".join(entry.categories)
        base_output[key] = entry
   
    output = deepcopy(base_output) 
    for entry_key, entry in base_output.items():
        categories = entry_key.split(":")
        last_category = entry_key
        for i in range(1, len(categories)):
            parent_categories = categories[:-i]
            category_key = ":".join(parent_categories)
            if category_key in output:
                output[category_key].amount += entry.amount
                if last_category not in output[category_key].children:
                    output[category_key].children.append(last_category)
                last_category = category_key
                continue
            output[category_key] = Entry(parent_categories,
                                         entry.is_income,
                                         True, 
                                         entry.amount,
                                         [last_category]
                                         )
            last_category = category_key
    return output

def generate_diff_dict(category_dicts: list[dict[str, Entry]]) -> dict[str, list[Union[Entry, None]]]:
    diff_dict = {}
    for dict_i, category_dict in enumerate(category_dicts): 
        for key, entry in category_dict.items():
            if key not in diff_dict:
                diff_dict[key] = [None] * len(category_dicts)
            diff_dict[key][dict_i] = entry
    return diff_dict

def entries_to_report_data(entries: list[Entry]) -> ReportData:
    income_entries = list(filter(lambda e: e.is_income, entries))
    expense_entries = list(filter(lambda e: not e.is_income, entries))
    
    report_data = ReportData(
        entries,
        calculate_total(income_entries),
        generate_category_dict(income_entries),
        calculate_total(expense_entries),
        generate_category_dict(expense_entries),
        calculate_total(entries)
    )
    return report_data



import shutil
import subprocess
import csv
import io
from typing import Union
from minibudget.model import Entry
from decimal import Decimal

def beancount_to_entry(record: dict, currency: str) -> Union[Entry, None]:
    amount = 0
    found_currency = False
    for key, val in record.items():
        if len(val) == 0: continue
        if currency in key: 
            # since it's double-entry
            # income and expenses are reversed
            amount = Decimal(val) * -1
            found_currency = True
            break
    if not found_currency:
        return None
    categories = record["account"].split(":")
    return Entry(
        categories=categories[1:],
        is_calculated=False,
        is_income=(categories[0] == "Income"),
        # the linter hates this, but it's succinct
        amount = int(amount * (Decimal('10') ** -amount.as_tuple()[2])),
        children=[]
    )

def beancount(file: str, currency: str, start: str, end: str) -> list[Entry]:
    # Validate bean-query is in PATH
    if shutil.which("bean-query") == None:
        raise EnvironmentError("bean-query could not be found. Please make sure it's installed in your environment and try again.")
    date_part = ""
    date_inner = []
    if start != None:
        date_inner.append(f"date >= {start}")
    if end != None:
        date_inner.append(f"date <= {end}")
    if start != None or end != None:
        date_part = f" and ( { " and ".join(date_inner) } )"
    # Issue query to get chart of accounts
    output = subprocess.run(["bean-query",
                             file,
                             f"select account, sum(position) where (account ~ 'Expenses' or account ~ 'Income' ) {date_part} group by account order by account",
                             "--format",
                             "csv",
                             "-m"], 
                            capture_output=True)
    output_as_csv = csv.DictReader(io.StringIO(output.stdout.decode('utf-8')))
    # Filter chart of accounts to the currency we care about
    entries = []
    for record in output_as_csv:
        entry = beancount_to_entry(record, currency)
        if entry is None: continue
        entries.append(entry)
    # Return accounts in minibudget format
    return entries

def entry_list_to_string(entry_list: list[Entry], width = 80):
    str_list = []
    for entry in entry_list:
        left = ""
        right = ""
        if entry.is_income:
            left += "+ "
        else:
            left += "- "
        left += ":".join(entry.categories)
        right += str(abs(entry.amount))
        spacer = " " * (width - (len(left) + len(right)))
        str_list.append(f"{left}{spacer}{right}")
    return "\n".join(str_list)



from minibudget.model import Entry
import sys

def tokenise_line(ln: str) -> list[str]:
    delimiter = "\""
    separator = " \n"
    buffer = ""
    fields = []
    delimited = False
    for char in ln:
        if char == delimiter:
            delimited = not delimited
        elif char in separator and not delimited and len(buffer) > 0:
            fields.append(buffer)
            buffer = ""
        elif (char not in separator) or delimited:
            buffer += char
    return fields

def line(ln: str) -> Entry:
    fields = tokenise_line(ln)
    
    if fields[0] not in "+-":
        raise ValueError("Only + and - are valid entries for a start of line.")

    is_expense = fields[0] == "-"

    categories = fields[1].split(":")
    amount = int(fields[2])

    if is_expense:
        amount = amount * -1

    return Entry(
        categories=categories,
        is_income=not is_expense,
        is_calculated=False,
        amount=amount,
        children=[]
    )

def budget(filename: str):
    entries: list[Entry] = []
    with open(filename) as f:
        for i, ln in enumerate(f):
            try:
                entry = line(ln)
                entries.append(entry)
            except Exception as err:
                print(err,file=sys.stderr)
                print(f"Couldn't parse line {i}; {ln}.", file=sys.stderr) 
    return entries

